# Licensed under the Apache License: http://www.apache.org/licenses/LICENSE-2.0
# For details: https://github.com/nedbat/coveragepy/blob/master/NOTICE.txt

"""Coverage.py's main entry point."""

import sys
from coverage.cmdline import main
sys.exit(main())
